import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * Class implementing known-plaintext and adaptive chosen-plaintext
 * cryptanalysis of Terry Ritter's k-Bit Block Cipher.
 * <p>
 * For a brief description of (mod 2)(mod p) polynomial arithmetic, see
 * http://www.io.com/~ritter/GLOSSARY.HTM#Mod2Polynomial
 */
public final class Solver {
    /** Constant passed to 'solve' for a known-plaintext attack. */
    public static final int KNOWN_PLAINTEXT = 0;

    /** Constant passed to 'solve' for an adaptive chosen-plaintext attack. */
    public static final int ADAPTIVE_CHOSEN_PLAINTEXT = 1;

    /** The PrintWriter that output is to be sent to. */
    private PrintWriter out;

    /** Whether to display full output. */
    boolean verbose;

    /** Whether to optimise for lower memory usage. */
    boolean lowmemory;

    /** Number of elements in the finite field (or entries in each table). */
    private int m;

    /** The primitive polynomial. */
    private int p;

    // Tables of known variables.
    private int[] TA;
    private int[] TB;
    private int[] ITX;
    private int[] ITY;

    /** A flag indicating an unknown value. */
    private static final int UNKNOWN = -1;

    /** Number of unknown variables. */
    private int unknowncount;

    /** Number of operations performed. */
    private int nops;

    /**
     * The set of plaintext/ciphertext pairs. (These are stored as
     * keys of a Hashtable only because there is no Set class in the
     * standard Java 1.1 API.)
     */
    private Hashtable textpairs = new Hashtable();

    /** The source of known/chosen plaintexts. */
    private TextSource source;

    /**
     * Creates a Solver with the given source of known/chosen plaintexts.
     */
    public Solver(TextSource source) {
        this.source = source;
        m = source.getM();
        p = source.getP();
        initTables();
    }

    public void reset() {
        TA = new int[m];
        TB = new int[m];
        ITX = new int[m];
        ITY = new int[m];

        for (int i = 0; i < m; i++) {
            TA[i] = TB[i] = ITX[i] = ITY[i] = UNKNOWN;
        }
        unknowncount = m*4;
        nops = 0;
    }

    // Precomputed tables to multiply and divide by 2 or 3.
    private int[] MULTIPLY_BY_2;
    private int[] MULTIPLY_BY_3;
    private int[] DIVIDE_BY_2;
    private int[] DIVIDE_BY_3;

    /** Initialise the tables used to multiply and divide by 2 or 3. */
    private void initTables() {
        MULTIPLY_BY_2 = new int[m];
        MULTIPLY_BY_3 = new int[m];
        DIVIDE_BY_2 = new int[m];
        DIVIDE_BY_3 = new int[m];

        for (int x = 0; x < m; x++) {
            int x2 = multiplyBy2(x);
            int x3 = multiplyBy3(x);

            MULTIPLY_BY_2[x] = x2;
            MULTIPLY_BY_3[x] = x3;
            DIVIDE_BY_2[x2] = x;
            DIVIDE_BY_3[x3] = x;
        }
    }

    private int multiplyBy2(int x) {
        int t = x << 1;
        if ((t & m) != 0)
            t ^= p;
        return t;
    }

    private int multiplyBy3(int x) {
        int t = x << 1;
        if ((t & m) != 0)
            t ^= p;
        return t^x;
    }

    private int multiply(int x, int y) {
        if (x == 0 || y == 0) return 0;

        int t = multiply(x >> 1, y) << 1;
        if ((x & 1) != 0)
            t ^= y;
        if ((t & m) != 0)
            t ^= p;
        return t;
    }

    /** Sets the PrintWriter that output is to be sent to. */
    public void setOutput(PrintWriter pw) { out = pw; }

    /** Sets whether verbose output is printed. */
    public void setVerbose(boolean flag) { verbose = flag; }

    /**
     * Sets whether to optimise for lower memory usage (at the expense of
     * requiring more plaintext).
     */
    public void setLowMemory(boolean flag) { lowmemory = flag; }

    /** Returns true if the cipher has been solved. */
    public boolean isSolved() { return unknowncount == 0; }

    /** Returns the number of table entries not yet found. */
    public int getUnknownCount() { return unknowncount; }

    /** Returns the number of operations performed. */
    public int getNOperations() { return nops; }

    /** Performs the type of attack given by the attacktype parameter. */
    public void solve(int attacktype) {
        switch (attacktype) {
          case KNOWN_PLAINTEXT:
            knownPlaintextAttack();
            break;
          case ADAPTIVE_CHOSEN_PLAINTEXT:
            adaptiveChosenPlaintextAttack();
            break;
          default:
            throw new IllegalArgumentException("invalid attack type (" +
                attacktype + ")");
        }
    }

    /** Performs a known-plaintext attack. */
    public void knownPlaintextAttack() {
        TextPair buf = lowmemory ? new TextPair() : null;
        reset();
        int prevcount = unknowncount;

        nops++;
        TextPair text1 = source.makeKnownPlaintext(buf);
        int ina1 = text1.getInA();
        int inb1 = text1.getInB();
        int outx1 = text1.getOutX();
        int outy1 = text1.getOutY();

        TA[ina1] = 0;
        TB[inb1] = 0;
        ITX[outx1] = 0;
        ITY[outy1] = 0;
        unknowncount -= 4;

        if (verbose) {
            out.println(text1 + " [first]");
            out.println("  => TA[" + ina1 + "] = " + TA[ina1]);
            out.println("  => TB[" + inb1 + "] = " + TB[inb1]);
            out.println("  => ITX[" + outx1 + "] = " + ITX[outx1]);
            out.println("  => ITY[" + outy1 + "] = " + ITY[outy1]);
        }

        int ina2, inb2;
        outer: while (true) {
            nops++;
            TextPair text2 = source.makeKnownPlaintext(buf);
            ina2 = text2.getInA();
            inb2 = text2.getInB();

            if (verbose) out.println(text2 + " [second]");

            if (ina2 != ina1 && inb2 != inb1) {
                if (text2.getOutX() == outx1) {
                    TA[ina2] = 1;
                    TB[inb2] = DIVIDE_BY_2[3];
                    break outer;
                }
                if (text2.getOutY() == outy1) {
                    TA[ina2] = 1;
                    TB[inb2] = DIVIDE_BY_3[2];
                    break outer;
                }
            }
            if (!lowmemory) textpairs.put(text2, text2);
        }

        if (verbose) {
            out.println("  => TA[" + ina2 + "] = " + TA[ina2]);
            out.println("  => TB[" + inb2 + "] = " + TB[inb2]);
        }
        unknowncount -= 2;

        mainloop: while (unknowncount > 0) {
            nops++;
            if (!lowmemory) {
                while (unknowncount < prevcount) {
                    nops++;
                    prevcount = unknowncount;
                    Enumeration e = textpairs.keys();

                    while (e.hasMoreElements()) {
                        nops++;
                        TextPair text = (TextPair) e.nextElement();
                        processTextPair(text, null);
                        if (unknowncount == 0)
                            break mainloop;
                    }
                }
            }
            TextPair text = source.makeKnownPlaintext(buf);
            if (!lowmemory) textpairs.put(text, text);
            processTextPair(text, null);
        }

        textpairs.clear();
    }

    /** Performs an adaptive chosen-plaintext attack. */
    public void adaptiveChosenPlaintextAttack() {
        TextPair buf = lowmemory ? new TextPair() : null;
        reset();
        int prevcount = unknowncount;

        nops++;
        TextPair text1 = source.makeChosenPlaintext(0, 0, buf);
        if (!lowmemory) textpairs.put(text1, text1);
        int outx1 = text1.getOutX();
        int outy1 = text1.getOutY();

        TA[0] = 0;
        TB[0] = 0;
        ITX[outx1] = 0;
        ITY[outy1] = 0;
        unknowncount -= 4;

        if (verbose) {
            out.println(text1 + " [first]");
            out.println("  => TA[0] = " + TA[0]);
            out.println("  => TB[0] = " + TB[0]);
            out.println("  => ITX[" + outx1 + "] = " + ITX[outx1]);
            out.println("  => ITY[" + outy1 + "] = " + ITY[outy1]);
        }

        int inb2;
        for (inb2 = 1; inb2 < m; inb2++) {
            nops++;
            TextPair text2 = source.makeChosenPlaintext(1, inb2, buf);
            if (!lowmemory) textpairs.put(text2, text2);

            if (verbose) out.println(text2 + " [second]");

            if (text2.getOutX() == outx1) {
                TA[1] = 1;
                TB[inb2] = DIVIDE_BY_2[3];
                break;
            }
            if (text2.getOutY() == outy1) {
                TA[1] = 1;
                TB[inb2] = DIVIDE_BY_3[2];
                break;
            }
        }

        if (verbose) {
            out.println("  => TA[1] = " + TA[1]);
            out.println("  => TB[" + inb2 + "] = " + TB[inb2]);
        }
        unknowncount -= 2;

        QueueEntry qhead;
        QueueEntry qtail = new QueueEntry(0, true,                   // ina == 0
                           new QueueEntry(0, false,                  // inb == 0
                           new QueueEntry(1, true,                   // ina == 1
                           (qhead = new QueueEntry(inb2, false, null // inb == inb2
                           )))));

        mainloop: while (unknowncount > 0 && qtail != null) {
            nops++;
            int n = qtail.n;
            boolean isrow = qtail.isrow;
            qtail = qtail.next;

            for (int i = 0; i < m; i++) {
                nops++;
                int ina = isrow ? n : i;
                int inb = isrow ? i : n;

                TextPair text = lowmemory ? null
                                          : (TextPair) textpairs.get(new TextPair(ina, inb));
                if (text == null) {
                    text = source.makeChosenPlaintext(ina, inb, buf);
                    if (!lowmemory) textpairs.put(text, text);
                }
                qhead = processTextPair(text, qhead);
                if (unknowncount == 0)
                    break mainloop;
            }
        }

        textpairs.clear();
    }

    static class QueueEntry {
        int n;
        boolean isrow;
        QueueEntry next;
        QueueEntry(int n, boolean isrow, QueueEntry next) {
            this.n = n;
            this.isrow = isrow;
            this.next = next;
        }
        QueueEntry add(int n, boolean isrow) {
            next = new QueueEntry(n, isrow, null);
            return next;
        }
    }

    /**
     * Processes the given TextPair, by filling in any table entries
     * that can be inferred from it.
     * <p>
     * If qhead is non-null and a TA or TB entry is found, a QueueEntry
     * for the corresponding row or column is added to the queue, and the
     * new queue head is returned.
     */
    private QueueEntry processTextPair(TextPair text, QueueEntry qhead) {
        int ina = text.getInA();
        int inb = text.getInB();
        int outx = text.getOutX();
        int outy = text.getOutY();
        int a = TA[ina];
        int b = TB[inb];
        int x = ITX[outx];
        int y = ITY[outy];

        // 3a + 2b + x = 0
        // 2a + 3b + y = 0

        if (verbose) out.println(text);

        if (b != UNKNOWN) {
            if (a == UNKNOWN) {
                if (x != UNKNOWN) {
                    TA[ina] = DIVIDE_BY_3[MULTIPLY_BY_2[b] ^ x];
                    if (verbose) out.println("  => TA[" + ina + "] = " + TA[ina]);
                    if (qhead != null) qhead = qhead.add(ina, true);
                    unknowncount--;
                } else if (y != UNKNOWN) {
                    TA[ina] = DIVIDE_BY_2[MULTIPLY_BY_3[b] ^ y];
                    if (verbose) out.println("  => TA[" + ina + "] = " + TA[ina]);
                    if (qhead != null) qhead = qhead.add(ina, true);
                    unknowncount--;
                }
            } else {
                if (x == UNKNOWN) {
                    ITX[outx] = MULTIPLY_BY_3[a] ^ MULTIPLY_BY_2[b];
                    if (verbose) out.println("  => ITX[" + outx + "] = " + ITX[outx]);
                    unknowncount--;
                }
                if (y == UNKNOWN) {
                    ITY[outy] = MULTIPLY_BY_2[a] ^ MULTIPLY_BY_3[b];
                    if (verbose) out.println("  => ITY[" + outy + "] = " + ITY[outy]);
                    unknowncount--;
                }
            }
        } else if (a != UNKNOWN) {
            if (x != UNKNOWN) {
                TB[inb] = DIVIDE_BY_2[MULTIPLY_BY_3[a] ^ x];
                if (verbose) out.println("  => TB[" + inb + "] = " + TB[inb]);
                if (qhead != null) qhead = qhead.add(inb, false);
                unknowncount--;
            } else if (y != UNKNOWN) {
                TB[inb] = DIVIDE_BY_3[MULTIPLY_BY_2[a] ^ y];
                if (verbose) out.println("  => TB[" + inb + "] = " + TB[inb]);
                if (qhead != null) qhead = qhead.add(inb, false);
                unknowncount--;
            }
        } else if (x != UNKNOWN && y != UNKNOWN) {
            TA[ina] = MULTIPLY_BY_3[x] ^ MULTIPLY_BY_2[y];
            TB[inb] = MULTIPLY_BY_2[x] ^ MULTIPLY_BY_3[y];
            if (verbose) out.println("  => TA[" + ina + "] = " + TA[ina]);
            if (verbose) out.println("  => TB[" + inb + "] = " + TB[inb]);
            if (qhead != null) qhead = qhead.add(ina, true);
            if (qhead != null) qhead = qhead.add(inb, false);
            unknowncount -= 2;
        }

        return qhead;
    }

    public void checkSolution(int[] origTA, int[] origTB,
                              int[] origTX, int[] origTY) {
        // f(x) = sx + t
        // g(x) = sx + u
        // u(x) = (x + 3t + 2u)/s
        // v(x) = (x + 2t + 3u)/s
        // TA = f.origTA
        // TB = g.origTB
        // TX = origTX.u
        // TY = origTY.v
        //
        // Check whether there is a solution for s != 0, t, and u.

        int i;
        for (i = 0; i < m; i++) {
            if (origTA[i] == 0) break;
        }
        if (i >= m) throw new Error("No i such that origTA[i] == 0");
        int t = TA[i];

        for (i = 0; i < m; i++) {
            if (origTA[i] == 1) break;
        }
        if (i >= m) throw new Error("No i such that origTA[i] == 1");
        int s = TA[i] ^ t;
        if (s == 0) throw new Error("s == 0");

        for (i = 0; i < m; i++) {
            if (origTB[i] == 0) break;
        }
        if (i >= m) throw new Error("No i such that origTB[i] == 0");
        int u = TB[i];

        int xc = MULTIPLY_BY_3[t] ^ MULTIPLY_BY_2[u]; // 3t + 2u
        int yc = MULTIPLY_BY_2[t] ^ MULTIPLY_BY_3[u]; // 2t + 3u

        int[] TX = getTX();
        int[] TY = getTY();
        int[] IorigTX = invert(origTX);
        int[] IorigTY = invert(origTY);

        for (i = 0; i < m; i++) {
            if (TX[i] == UNKNOWN || TY[i] == UNKNOWN ||
                (multiply(s, origTA[i]) ^ t) != TA[i] ||
                (multiply(s, origTB[i]) ^ u) != TB[i] ||
                (multiply(s, IorigTX[TX[i]]) ^ xc) != i ||
                (multiply(s, IorigTY[TY[i]]) ^ yc) != i)
                throw new Error("Solution is incorrect (possibly p is not irreducible?)");
        }

        out.println("Solution is equivalent to the original key.");
    }

    public void dump() {
        out.println("Dump of known variables (numbers are in hex):");
        int w = (source.getK()+7)/8;

        int[] TX = getTX();
        int[] TY = getTY();
        for (int i = 0; i < m; i++) {
            out.println("  TA[" + Hex.toString(i, w) + "] = " + Hex.toString(TA[i], w) +
                        ", TB[" + Hex.toString(i, w) + "] = " + Hex.toString(TB[i], w) +
                        ", TX[" + Hex.toString(i, w) + "] = " + Hex.toString(TX[i], w) +
                        ", TY[" + Hex.toString(i, w) + "] = " + Hex.toString(TY[i], w));
        }
    }

    public int[] getTA() { return TA; }
    public int[] getTB() { return TB; }
    public int[] getITX() { return ITX; }
    public int[] getITY() { return ITY; }
    public int[] getTX() { return invert(ITX); }
    public int[] getTY() { return invert(ITY); }

    /** Returns the inverse of the given permutation table. */
    public int[] invert(int[] tab) {
        int[] inv = new int[m];
        for (int i = 0; i < m; i++)
            inv[i] = UNKNOWN;

        for (int i = 0; i < m; i++)
            if (tab[i] != UNKNOWN) inv[tab[i]] = i;

        return inv;
    }
}
