import java.util.Hashtable;

/**
 * The k-Bit Block Cipher in Java. For research analysis; not a useful
 * cipher.
 *
 * @author Terry Ritter (C implementation)
 * @author David Hopwood
 */
public class kBit {
    private int k;
    private int m; // 1 << (k/2)
    private int p;

    public int getK() { return k; }
    public int getM() { return m; }
    public int getP() { return p; }

    // Precomputed tables to multiply by 2 or 3.
    private int[] MULTIPLY_BY_2;
    private int[] MULTIPLY_BY_3;

    /** Initialise the tables used to multiply by 2 or 3. */
    private void initTables() {
        MULTIPLY_BY_2 = new int[m];
        MULTIPLY_BY_3 = new int[m];

        for (int x = 0; x < m; x++) {
            MULTIPLY_BY_2[x] = multiplyBy2(x);
            MULTIPLY_BY_3[x] = multiplyBy3(x);
        }
    }

    private int multiplyBy2(int x) {
        int t = x << 1;
        if ((t & m) != 0)
            t ^= p;
        return t;
    }

    private int multiplyBy3(int x) {
        int t = x << 1;
        if ((t & m) != 0)
            t ^= p;
        return t^x;
    }

    // keyed tables
    private int TA[];
    private int TB[];
    private int TX[];
    private int TY[];

    public int[] getTA() { return TA; }
    public int[] getTB() { return TB; }
    public int[] getTX() { return TX; }
    public int[] getTY() { return TY; }

    private static void initTab(int tab[]) {
        for (int i = 0; i < tab.length; i++) {
            tab[i] = i;
        }
    }

    /**
     * Shuffle the table. The table length must be a power of 2.
     */
    private static void shufTab(int tab[]) {
        int mask = tab.length - 1;
        int nxtmsk, N, rand, by;

        do {
            nxtmsk = mask >> 1;

            for (N = mask; N > nxtmsk; N--) {
                rand = RNG.rand0thruN(mask, N);
                by = tab[N];
                tab[N] = tab[rand];
                tab[rand] = by;
            }

            mask = nxtmsk;
        } while (mask != 0);
    }


    // object state ciphering

    private int InA, InB, A, B, X, Y, OutX, OutY;

    public void setInput(int a, int b) { InA = a; InB = b; mix(); }
    public int getInA() { return InA; }
    public int getInB() { return InB; }
    public int getA() { return A; }
    public int getB() { return B; }
    public int getX() { return X; }
    public int getY() { return Y; }
    public int getOutX() { return OutX; }
    public int getOutY() { return OutY; }

    private void mix() {
        A = TA[InA];
        B = TB[InB];
        X = MULTIPLY_BY_3[A] ^ MULTIPLY_BY_2[B];
        Y = MULTIPLY_BY_2[A] ^ MULTIPLY_BY_3[B];
        OutX = TX[X];
        OutY = TY[Y];
    }


    // functional ciphering

    public int mixX(int ina, int inb) {
        return TX[MULTIPLY_BY_3[TA[ina]] ^ MULTIPLY_BY_2[TB[inb]]];
    }

    public int mixY(int ina, int inb) {
        return TY[MULTIPLY_BY_2[TA[ina]] ^ MULTIPLY_BY_3[TB[inb]]];
    }


    /**
     * Constructor for a k-Bit cipher. If trivial is true, the keying tables
     * are unshuffled (i.e. always the identity permutation). Otherwise, they
     * are set to a random permutation.
     *
     * @param k         the block size in bits (even, and from 4 to 60 inclusive)
     * @param p         the irreducable polynomial used for the BBM (from
     *                    m to m*2-1 inclusive, where m = 2<sup>k/2</sup>)
     * @param trivial   to force the use of unshuffled tables
     * @exception IllegalArgumentException if k or p is invalid.
     */
    public kBit(int k, int p, boolean trivial) {
        if (k < 4 || k > 60 || k%2 != 0)
            throw new IllegalArgumentException("k < 4 || k > 60 || k%2 != 0");

        m = 1 << (k/2);
        if (p < m || p >= m*2)
            throw new IllegalArgumentException("p < m || p >= m*2");

        this.k = k;
        this.p = p;
        initTables();

        // initialise the 4 keying tables
        TA = new int[m]; initTab(TA);
        TB = new int[m]; initTab(TB);
        TX = new int[m]; initTab(TX);
        TY = new int[m]; initTab(TY);

        if (!trivial) {
            shufTab(TA);
            shufTab(TB);
            shufTab(TX);
            shufTab(TY);
        }
    }
}
