import java.io.PrintWriter;

/**
 * Demonstration of known-plaintext and adaptive chosen-plaintext attacks
 * against Terry Ritter's k-Bit Block Cipher.
 * <p>
 * Run with no arguments for options.
 *
 * @author David Hopwood
 */
class kBitDemo {
    private static final String usage =
        "\n" +
        "Usage: java kBitDemo [-verbose] [-trivial] [-showtables] [-lowmemory]\n" +
        "                     [-known|-adaptive] <k> <p>\n" +
        "\n" +
        "where <k> is the block size in bits (even, and between 4 and 60)\n" +
        "      <p> is the irreducible polynomial used for the Balanced\n" +
        "          Block Mixing (a number from m to 2*m-1, where m = 2^(k/2))\n" +
        "\n" +
        "Options:\n" +
        "      -verbose     print more information\n" +
        "      -trivial     force tables to the identity permutation\n" +
        "      -showtables  show table of outputs, and key tables\n" +
        "      -lowmemory   optimise for lower memory usage, at the expense\n" +
        "                     of requiring more plaintext\n" +
        "      -known       perform a known-plaintext attack (default)\n" +
        "      -adaptive    perform an adaptive chosen-plaintext attack\n";

    /** The PrintWriter that output is to be sent to. */
    private PrintWriter out;

    public static void main(String[] args) {
        new kBitDemo().run(new PrintWriter(System.out, true), args);
    }

    public void run(PrintWriter out, String[] args) {
        this.out = out;
        boolean verbose = false, trivial = false,
                showtables = false, lowmemory = false;
        int attacktype = Solver.KNOWN_PLAINTEXT;

        int i;
        for (i = 0; i < args.length; i++) {
            if (args[i].equals("-verbose"))
                verbose = true;
            else if (args[i].equals("-trivial"))
                trivial = true;
            else if (args[i].equals("-showtables"))
                showtables = true;
            else if (args[i].equals("-lowmemory"))
                lowmemory = true;
            else if (args[i].equals("-known"))
                attacktype = Solver.KNOWN_PLAINTEXT;
            else if (args[i].equals("-adaptive"))
                attacktype = Solver.ADAPTIVE_CHOSEN_PLAINTEXT;
            else
                break;
        }
        if (args.length < i+2) {
            out.println(usage);
            return;
        }
        try {
            demo(Integer.parseInt(args[i]), Integer.parseInt(args[i+1]),
                verbose, trivial, showtables, lowmemory, attacktype);
        } catch (Exception e) {
            out.println(usage);
            e.printStackTrace();
        }
    }

    private void demo(int k, int p, boolean verbose,
                      boolean trivial, boolean showtables,
                      boolean lowmemory, int attacktype) {

        kBit cipher = new kBit(k, p, trivial);

        int m = cipher.getM();
        int w = (k+7)/8;

        int[] TA = cipher.getTA();
        int[] TB = cipher.getTB();
        int[] TX = cipher.getTX();
        int[] TY = cipher.getTY();

        if (showtables) {
            if (w > 1) {
                out.println("Output table is too large to display.");
            } else {
                int ina, inb, outx, outy;
                boolean seenXinThisRow[], seenYinThisRow[],
                        seenXinCol[][], seenYinCol[][];

                // show every possible plaintext and ciphertext, and check
                // that no X or Y output is repeated within a row or column.

                // print the inb column labels
                String spacer = Hex.spacer(' ', w);
                out.print("  " + spacer);
                for (inb = 0; inb < m; inb++) {
                    out.print(spacer + "  " + Hex.toString(inb, w));
                }

                seenXinCol = new boolean[m][m]; // all false
                seenYinCol = new boolean[m][m]; // all false

                for (ina = 0; ina < m; ina++) {
                    seenXinThisRow = new boolean[m]; // all false
                    seenYinThisRow = new boolean[m]; // all false

                    // start row with ina label
                    out.print("\n  " + Hex.toString(ina, w));

                    for (inb = 0; inb < m; inb++) {
                        cipher.setInput(ina, inb);
                        outx = cipher.getOutX();
                        outy = cipher.getOutY();

                        out.print("  " + Hex.toString(outx, outy, w));

                        if (seenXinThisRow[outx] || seenXinCol[inb][outx])
                            throw new Error("X output was repeated");

                        seenXinThisRow[outx] = seenXinCol[inb][outx] = true;

                        if (seenYinThisRow[outy] || seenYinCol[inb][outy])
                            throw new Error("Y output was repeated");

                        seenYinThisRow[outy] = seenYinCol[inb][outy] = true;
                    }
                }
                out.println("\n");
            }

            out.println("Dump of key tables (numbers are in hex):");
            for (int i = 0; i < m; i++) {
                out.println("  TA[" + Hex.toString(i, w) + "] = " + Hex.toString(TA[i], w) +
                            ", TB[" + Hex.toString(i, w) + "] = " + Hex.toString(TB[i], w) +
                            ", TX[" + Hex.toString(i, w) + "] = " + Hex.toString(TX[i], w) +
                            ", TY[" + Hex.toString(i, w) + "] = " + Hex.toString(TY[i], w));
            }
            out.println();
        }

        TextSource source = new TextSource(cipher);
        Solver solver = new Solver(source);
        solver.setOutput(out);
        solver.setVerbose(verbose);
        solver.setLowMemory(lowmemory);

        long start = System.currentTimeMillis();
        solver.solve(attacktype);

        long time = System.currentTimeMillis() - start;

        out.println("Used " + source.getNChosen() + " chosen and " +
            source.getNKnown() + " known plaintexts.");
        out.println("Cryptanalysis was " +
            (solver.isSolved() ? "" : "not ") + "completed after " +
            (verbose ? "" : ((time/1000.0d) + " seconds and ")) +
            solver.getNOperations() + " operations.");
        out.println();

        solver.dump();
        out.println();
        solver.checkSolution(TA, TB, TX, TY);
    }
}
